# tests/test_cache_metrics.py

import unittest

from kinto.core.testing import skip_if_no_memcached, skip_if_no_postgresql, skip_if_no_prometheus

from .support import BaseWebTest


@skip_if_no_prometheus
class BaseCacheMetricsTest(BaseWebTest):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["includes"] += "\nkinto.plugins.prometheus"
        return settings

    def setUp(self):
        super().setUp()
        from kinto.plugins.prometheus import reset_registry

        reset_registry()

    def test_cache_miss_and_hit_counters(self):
        # Miss: key does not exist.
        self.cache.get("unknown-key")

        # Hits: store then retrieve twice.
        self.cache.set("foo", {"bar": 42}, ttl=30)
        self.cache.get("foo")
        self.cache.get("foo")

        # Fetch the raw Prometheus metrics.
        resp = self.app.get("/__metrics__")

        # We should see 1 miss and 2 hits.
        self.assertIn("kinto_cache_misses_total 1.0", resp.text)
        self.assertIn("kinto_cache_hits_total 2.0", resp.text)


class MemoryCacheMetricsTest(BaseCacheMetricsTest, unittest.TestCase):
    """Run cache metrics tests with default (memory) backend."""

    pass


@skip_if_no_memcached
class MemcachedCacheMetricsTest(BaseCacheMetricsTest, unittest.TestCase):
    """Run cache metrics tests with Memcached cache backend."""

    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        # Switch to Memcached backend
        settings["cache_backend"] = "kinto.core.cache.memcached"
        settings["cache_hosts"] = "127.0.0.1:11211"
        return settings


@skip_if_no_postgresql
class PostgreSQLCacheMetricsTest(BaseCacheMetricsTest, unittest.TestCase):
    """Run cache metrics tests with PostgreSQL cache backend."""

    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        # Switch to PostgreSQL backend
        settings["cache_backend"] = "kinto.core.cache.postgresql"
        settings["cache_url"] = "postgresql://postgres:postgres@localhost:5432/testdb"
        return settings



import colander
from pyramid.httpexceptions import HTTPForbidden

from kinto.core import Service, resource, utils
from kinto.core.resource.schema import ErrorResponseSchema
from kinto.core.resource.viewset import ViewSet
from kinto.core.storage import Filter, Sort
from kinto.core.utils import instance_uri


class HistorySchema(resource.ResourceSchema):
    user_id = colander.SchemaNode(colander.String())
    uri = colander.SchemaNode(colander.String())
    action = colander.SchemaNode(colander.String())
    date = colander.SchemaNode(colander.String())
    resource_name = colander.SchemaNode(colander.String())
    bucket_id = colander.SchemaNode(colander.String(), missing=colander.drop)
    collection_id = colander.SchemaNode(colander.String(), missing=colander.drop)
    group_id = colander.SchemaNode(colander.String(), missing=colander.drop)
    record_id = colander.SchemaNode(colander.String(), missing=colander.drop)
    target = colander.SchemaNode(colander.Mapping())

    class Options:
        preserve_unknown = False


# Add custom OpenAPI tags/operation ids
plural_get_arguments = getattr(ViewSet, "plural_get_arguments", {})
plural_delete_arguments = getattr(ViewSet, "plural_delete_arguments", {})

get_history_arguments = {
    "tags": ["History"],
    "operation_id": "get_history",
    **plural_get_arguments,
}
delete_history_arguments = {
    "tags": ["History"],
    "operation_id": "delete_history",
    **plural_delete_arguments,
}


@resource.register(
    name="history",
    plural_path="/buckets/{{bucket_id}}/history",
    object_path=None,
    plural_methods=("GET", "DELETE"),
    default_arguments={"tags": ["History"], **ViewSet.default_arguments},
    plural_get_arguments=get_history_arguments,
    plural_delete_arguments=delete_history_arguments,
)
class History(resource.Resource):
    schema = HistorySchema

    def get_parent_id(self, request):
        self.bucket_id = request.matchdict["bucket_id"]
        return instance_uri(request, "bucket", id=self.bucket_id)

    def _extract_filters(self):
        filters = super()._extract_filters()
        filters_str_id = []
        for filt in filters:
            if filt.field in ("record_id", "collection_id", "bucket_id"):
                if isinstance(filt.value, int):
                    filt = Filter(filt.field, str(filt.value), filt.operator)
            filters_str_id.append(filt)

        return filters_str_id


snapshot = Service(
    name="history_snapshot",
    path="/buckets/{bucket_id}/snapshot/collections/{collection_id}@{timestamp}",
    description="Reconstruct collection at given timestamp",
)


def timestamp_validator(request, **kwargs):
    """
    Validates that the timestamp is an integer.
    """
    timestamp = request.matchdict["timestamp"]
    try:
        if int(timestamp) < 0:
            raise ValueError
    except ValueError:
        request.errors.add("path", "timestamp", "Invalid timestamp %r" % timestamp)


class SnapshotPathSchema(colander.MappingSchema):
    bucket_id = colander.SchemaNode(colander.String())
    collection_id = colander.SchemaNode(colander.String())
    timestamp = colander.SchemaNode(colander.Integer())


class SnapshotSchema(colander.MappingSchema):
    path = SnapshotPathSchema()


snapshot_response_schemas = {
    "401": ErrorResponseSchema(description="The collection is not publicly readable."),
    "403": ErrorResponseSchema(description="No permission to read this collection."),
    "200": colander.SchemaNode(
        colander.Mapping(),
        description="Returns the records at the given timestamp.",
    ),
}


@snapshot.get(
    schema=SnapshotSchema(),
    validators=(timestamp_validator,),
    response_schemas=snapshot_response_schemas,
)
def get_snapshot(request):
    """Reconstructs the collection as it was at the given timestamp."""
    bucket_id = request.matchdict["bucket_id"]
    collection_id = request.matchdict["collection_id"]
    timestamp = int(request.matchdict["timestamp"])

    bucket_uri = instance_uri(request, "bucket", id=bucket_id)
    collection_uri = instance_uri(request, "collection", bucket_id=bucket_id, id=collection_id)

    # Check that user has read permission on the collection.
    # This is manual code, because we are outside the normal resource system.
    if not request.registry.permission.check_permission(
        request.prefixed_principals,
        [
            (bucket_uri, "read"),
            (bucket_uri, "write"),
            (collection_uri, "read"),
            (collection_uri, "write"),
        ],
    ):
        raise HTTPForbidden()

    # List all the records that have changed since the given timestamp.
    all_records = request.registry.storage.list_all(
        parent_id=collection_uri,
        resource_name="record",
        include_deleted=True,  # Include tombstones
    )

    unchanged_records = [
        r for r in all_records if r["last_modified"] <= timestamp and not r.get("deleted")
    ]
    changed_rids = [r["id"] for r in all_records if r["last_modified"] > timestamp]
    if not changed_rids:
        # No change after timestamp, return all records as-is.
        return {"data": sorted(unchanged_records, key=lambda r: r["last_modified"], reverse=True)}

    # History entries store the current version. We need to pick the most recent
    # entry before the timestamp for each record_id to obtain the records' state
    # before it was changed or deleted.
    history_entries = request.registry.storage.list_all(
        parent_id=bucket_uri,
        resource_name="history",
        filters=[
            Filter("resource_name", "record", utils.COMPARISON.EQ),
            Filter("collection_id", collection_id, utils.COMPARISON.EQ),
            Filter("record_id", changed_rids, utils.COMPARISON.IN),
            Filter("target.data.last_modified", timestamp, utils.COMPARISON.MAX),
        ],
        sorting=[Sort("last_modified", -1)],  # Most recent first
        # TODO: add storage option to keep only the latest entry per record_id
    )

    most_recent_entry = {}
    for entry in history_entries:
        rid = entry["record_id"]
        if rid not in most_recent_entry:
            most_recent_entry[rid] = entry

    # Records created after the timestamp (not existing in history) should not appear.
    # Records deleted or updated after the timestamp should be reverted to their most recent
    # version before the timestamp.
    result_records = unchanged_records
    for rid in changed_rids:
        if rid not in most_recent_entry:
            # Record was created after the timestamp, skip it.
            continue
        history_entry = most_recent_entry[rid]
        result_records.append(history_entry["target"]["data"])

    return {"data": sorted(result_records, key=lambda r: r["last_modified"], reverse=True)}
